/*
 * ppc.c
 */

#ifdef AMIGA
#ifdef __amigaos4__
#include <exec/memory.h>
#include <proto/exec.h>
static void cache_flush(void *addr, uint32 bytes)
{
  CacheClearE(addr, bytes, CACRF_ClearI | CACRF_ClearD);
}
#else
#include <powerpc/powerpc.h>
#include <powerpc/powerpc_protos.h>
static void cache_flush(void *addr, uint32 bytes)
{
  /* invalidating the whole instruction cache is fastest */
  SetCache(CACHE_ICACHEINV, 0, 0);

  /* flushing only a part of the data cache is fastest */
  SetCache(CACHE_DCACHEFLUSH, addr, bytes);
}
#endif
#else
#error cache_flush needs to be implemented
#endif

#include "ppc.h"

/* ppc stack and register handling */

static void ppc_push_int(int int_reg, int pos)
{
  OPCODE = stw(int_reg, 24 + (pos * 4), 1);
}

static void ppc_pop_int(int int_reg, int pos)
{
  OPCODE = lwz(int_reg, 24 + (pos * 4), 1);
}

static void ppc_push_double(int double_reg, int pos)
{
  OPCODE = stfd(double_reg, 24 + (32*4) + (pos * 8), 1);
}

static void ppc_pop_double(int double_reg, int pos)
{
  OPCODE = lfd(double_reg, 24 + (32*4) + (pos * 8), 1);
}

#define EX ((32*4) + (32*8))

static void ppc_save(void)
{
  /* SAVE all registers */
  int x;

  /* stack pointer */
  OPCODE = stwu(1, -(240 + EX), 1);

  /* link register */
  OPCODE = mflr(0);
  OPCODE = stw(0, 244 + EX, 1);

  /* floating point registers (64bit) */
  for (x=14; x<=31; x++) {
    OPCODE = stfd(x, 96 + EX + ((x - 14) * 8), 1);
  }

  /* integer registers (32bit) */
  for (x=14; x<=31; x++) {
    OPCODE = stw(x, 24 + EX + ((x - 14) * 4), 1);
  }
}

static void ppc_load(void)
{
  /* LOAD all registers */
  int x;

  /* integer registers (32bit) */
  for (x=14; x<=31; x++) {
    OPCODE = lwz(x, 24 + EX + ((x - 14) * 4), 1);
  }

  /* floating point registers (64bit) */
  for (x=14; x<=31; x++) {
    OPCODE = lfd(x, 96 + EX + ((x - 14) * 8), 1);
  }

  /* link register */
  OPCODE = lwz(0, 244 + EX, 1);
  OPCODE = mtlr(0);

  /* stack pointer */
  OPCODE = la(1, 240 + EX, 1);
}

#undef EX
