/*
 * test68k.c
 */

#undef SRCTEST
#define SRCTEST

#undef DSTTEST
#define DSTTEST

#undef FLAGTEST
#define FLAGTEST

#undef VECTORTEST
#define VECTORTEST

#include "generator.h"
#include "registers.h"

#include "cpu68k.h"
#include "diss68k.h"
extern int diss68k_gettext(t_ipc * ipc, char *text);
#include "mem68k.h"

#include "ppc.h"
#include "common.h"

#include <stdio.h>
#include <stdlib.h>

#include <proto/exec.h>
#include <powerpc/powerpc.h>
#include <powerpc/powerpc_protos.h>

#define PC (reg68k_pc) /* FIXME: be carefull */

#define DATAS (18)

uint32 byte_data[DATAS] = {
  0x000000ff,
  0x00000000,
  0x00000001,
  0x00000080,
  0x0000007f,
  0x000000aa,
  0x00000055,
  7,
  8,
  9,
  15,
  16,
  17,
  31,
  32,
  33,
  0x000000fe,
  0x00000002,
};

uint32 word_data[DATAS] = {
  0x0000ffff,
  0x00000000,
  0x00000001,
  0x00008000,
  0x00007f00,
  0x0000aaaa,
  0x00005555,
  7,
  8,
  9,
  15,
  16,
  17,
  31,
  32,
  33,
  0x0000fffe,
  0x00000002,
};

uint32 long_data[DATAS] = {
  0xffffffff,
  0x00000000,
  0x00000001,
  0x80000000,
  0x7f000000,
  0xaaaaaaaa,
  0x55555555,
  7,
  8,
  9,
  15,
  16,
  17,
  31,
  32,
  33,
  0xfffffffe,
  0x00000002,
};

uint32 srcdata, _srcdata, dstdata, _dstdata, outdata, cc, instr_error;
void *ptr;

#define TFLAG (reg68k_sr.sr_struct.t)
#define SFLAG (reg68k_sr.sr_struct.s)
#define XFLAG (reg68k_sr.sr_struct.x)
#define NFLAG (reg68k_sr.sr_struct.n)
#define ZFLAG (reg68k_sr.sr_struct.z)
#define VFLAG (reg68k_sr.sr_struct.v)
#define CFLAG (reg68k_sr.sr_struct.c)

#include "compile_flags.h"

/* compiled */

static uint32 regs68k[16], regsppc[16];
static char *regnames[16] = { "D0", "D1", "D2", "D3", "D4", "D5", "D6", "D7", "A0", "A1", "A2", "A3", "A4", "A5", "A6", "A7" };

#define SRCREG (28) /* MUST be non-volatile */
#define DSTREG (29)

#define BUFSIZE (16384)

static void *buf;

typedef void (*op_t)(void);

void compile_setup(void)
{
  buf = malloc(BUFSIZE);
}

void compile_cleanup(void)
{
  free(buf);
}

/* stuff */

void _compile_cc(t_iib *iib)
{
  int scale;
  op_t op;

  scale = 5; /* just test a value */

  memset(buf, 0, BUFSIZE);
  ptr = buf;

  OPCODE = stwu(1, -240, 1);
  OPCODE = mflr(0);
  OPCODE = stw(0, 244, 1);

  compile_cc(iib, 4, scale);

  OPCODE = rlwinm(4, 4, (32 - scale), 31, 31);

  OPCODE = lis(3, ((uint32)&cc >> 16));
  OPCODE = ori(3, 3, ((uint32)&cc & 0xffff));
  OPCODE = stw(4, 0, 3);

  OPCODE = lwz(0, 244, 1);
  OPCODE = mtlr(0);
  OPCODE = la(1, 240, 1);
  OPCODE = blr();

  /* Flush the cache */
  SetCache(CACHE_ICACHEINV, 0, 0);
  SetCache(CACHE_DCACHEFLUSH, 0, 0);

  /* call */
  op = (op_t)buf;
  op();
}

#define VFLAG_ADD_8(S, D, R) (((S^R) & (D^R))>>7)
#define VFLAG_ADD_16(S, D, R) (((S^R) & (D^R))>>15)
#define VFLAG_ADD_32(S, D, R) (((S^R) & (D^R))>>31)

#define VFLAG_SUB_8(S, D, R) (((S^D) & (R^D))>>7)
#define VFLAG_SUB_16(S, D, R) (((S^D) & (R^D))>>15)
#define VFLAG_SUB_32(S, D, R) (((S^D) & (R^D))>>31)

#define CXFLAG_NEGX_8(S, R) ((S|R)>>7)
#define CXFLAG_NEGX_16(S, R) ((S|R)>>15)
#define CXFLAG_NEGX_32(S, R) ((S|R)>>31)

#define VFLAG_NEGX_8(S, R) ((S&R)>>7)
#define VFLAG_NEGX_16(S, R) ((S&R)>>15)
#define VFLAG_NEGX_32(S, R) ((S&R)>>31)

/* portable */

void generate_cc(t_iib *iib)
{
  switch(iib->cc) {
  case 0:                                                     /*  T */
    cc = 1;
    break;
  case 1:                                                     /*  F */
    cc = 0;
    break;
  case 2:                                                     /* HI */
    cc = !(CFLAG || ZFLAG);
    break;
  case 3:                                                     /* LS */
    cc = CFLAG || ZFLAG;
    break;
  case 4:                                                     /* CC */
    cc = !CFLAG;
    break;
  case 5:                                                     /* CS */
    cc = CFLAG;
    break;
  case 6:                                                     /* NE */
    cc = !ZFLAG;
    break;
  case 7:                                                     /* EQ */
    cc = ZFLAG;
    break;
  case 8:                                                     /* VC */
    cc = !VFLAG;
    break;
  case 9:                                                     /* VS */
    cc = VFLAG;
    break;
  case 10:                                                    /* PL */
    cc = !NFLAG;
    break;
  case 11:                                                    /* MI */
    cc = NFLAG;
    break;
  case 12:                                                    /* GE */
    cc = (NFLAG == VFLAG);
    break;
  case 13:                                                    /* LT */
    cc = (NFLAG != VFLAG);
    break;
  case 14:                                                    /* GT */
    cc = !ZFLAG && (NFLAG == VFLAG);
    break;
  case 15:                                                    /* LE */
    cc = ZFLAG || (NFLAG != VFLAG);
    break;
  default:
    break;
  }
}

void generate_stdflag_n(t_iib *iib)
{
  switch(iib->size) {
  case sz_byte:
    NFLAG = ((sint8)outdata) < 0;
    break;
  case sz_word:
    NFLAG = ((sint16)outdata) < 0;
    break;
  case sz_long:
    NFLAG = ((sint32)outdata) < 0;
    break;
  default:
    break;
  }
}

void generate_stdflag_z(t_iib *iib)
{
  ZFLAG = !outdata;
}

/* clrflag */

void generate_clrflag_c(t_iib *iib)
{
  CFLAG = 0;
}

void generate_clrflag_v(t_iib *iib)
{
  VFLAG = 0;
}

void generate_clrflag_z(t_iib *iib)
{
  ZFLAG = 0;
}

void generate_clrflag_n(t_iib *iib)
{
  NFLAG = 0;
}

void generate_clrflag_x(t_iib *iib)
{
  XFLAG = 0;
}

void generate_clrflags(uint32 set)
{
  if (set & IIB_FLAG_C)
    generate_clrflag_c(NULL);
  if (set & IIB_FLAG_V)
    generate_clrflag_v(NULL);
  if (set & IIB_FLAG_Z)
    generate_clrflag_z(NULL);
  if (set & IIB_FLAG_N)
    generate_clrflag_n(NULL);
  if (set & IIB_FLAG_X)
    generate_clrflag_x(NULL);
}

/* setflag */

void generate_setflag_c(t_iib *iib)
{
  CFLAG = 1;
}

void generate_setflag_v(t_iib *iib)
{
  VFLAG = 1;
}

void generate_setflag_z(t_iib *iib)
{
  ZFLAG = 1;
}

void generate_setflag_n(t_iib *iib)
{
  NFLAG = 1;
}

void generate_setflag_x(t_iib *iib)
{
  XFLAG = 1;
}

void generate_setflags(uint32 set)
{
  if (set & IIB_FLAG_C)
    generate_setflag_c(NULL);
  if (set & IIB_FLAG_V)
    generate_setflag_v(NULL);
  if (set & IIB_FLAG_Z)
    generate_setflag_z(NULL);
  if (set & IIB_FLAG_N)
    generate_setflag_n(NULL);
  if (set & IIB_FLAG_X)
    generate_setflag_x(NULL);
}

void generate_subflag_c(t_iib *iib)
{
  /* C = (Sm && !Dm) || (Rm && !Dm) || (Sm && Rm)
     carry is performed as if both source and destination are unsigned,
     so this is simply if the source is greater than the destination - I
     have proven this to be the case using a truth table and the above
     motorola definition */
  CFLAG = srcdata > dstdata;
}

void generate_subflag_cx(t_iib *iib)
{
  /* C = (Sm && !Dm) || (Rm && !Dm) || (Sm && Rm)
     carry is performed as if both source and destination are unsigned,
     so this is simply if the source is greater than the destination - I
     have proven this to be the case using a truth table and the above
     motorola definition */
  XFLAG = CFLAG = srcdata > dstdata;
}

void generate_subxflag_cx(t_iib *iib)
{
  /* V = (Sm && !Dm) || (Rm && (!Dm || Sm)) */
  int Sm = 0, Dm = 0, Rm = 0;

  switch(iib->size) {
  case sz_byte:
    Sm = (sint8)srcdata < 0;
    Dm = (sint8)dstdata < 0;
    Rm = (sint8)outdata < 0;
    break;
  case sz_word:
    Sm = (sint16)srcdata < 0;
    Dm = (sint16)dstdata < 0;
    Rm = (sint16)outdata < 0;
    break;
  case sz_long:
    Sm = (sint32)srcdata < 0;
    Dm = (sint32)dstdata < 0;
    Rm = (sint32)outdata < 0;
    break;
  default:
    break;
  }

  XFLAG = CFLAG = (Sm && !Dm) || (Rm && (!Dm || Sm));
}

void generate_cmpaflag_c(t_iib *iib)
{
  /* see generate_subflag_c - this is just the same but with a sign extend
     on the source */
  CFLAG = (uint32)(sint32)(sint16)srcdata > dstdata;
}

void generate_subflag_v(t_iib *iib)
{
  /* V = (!Sm && Dm && !Rm) || (Sm && !Dm && Rm)
     overflow is performed as if both source and destination are signed,
     the only two condtions are if we've added too much to a +ve number
     or we've subtracted too much from a -ve number - I have proven the
     this to be the case using a truth table and the above motorola
     definition */
  /* the technique to implement the above formula is to make sure the sign
     of Sm != the sign of Dm, and the sign of Dm != the sign of Rm. */
  switch(iib->size) {
  case sz_byte:
    VFLAG = (((sint8)srcdata < 0) != ((sint8)dstdata < 0))
         && (((sint8)dstdata < 0) != ((sint8)outdata < 0));
    break;
  case sz_word:
    VFLAG = (((sint16)srcdata < 0) != ((sint16)dstdata < 0))
         && (((sint16)dstdata < 0) != ((sint16)outdata < 0));
    break;
  case sz_long:
    VFLAG = (((sint32)srcdata < 0) != ((sint32)dstdata < 0))
         && (((sint32)dstdata < 0) != ((sint32)outdata < 0));
    break;
  default:
    break;
  }
}

void generate_cmpaflag_v(t_iib *iib)
{
  /* see generate_subflag_v - this is just the sz_long version with a sign
     extend on the source */
  VFLAG = (((sint32)(sint16)srcdata < 0) != ((sint32)dstdata < 0))
       && (((sint32)dstdata < 0) != ((sint32)outdata < 0));
}

void generate_stdxflag_z(t_iib *iib)
{
  if (outdata) ZFLAG = 0;
}

void generate_addflag_cx(t_iib *iib)
{
  /* C = (Sm && Dm) || (!Rm && Dm) || (Sm && !Rm)
     carry is performed as if both source and destination are unsigned,
     so this is simply if the source is bigger than how much can be added
     to the destination without wrapping - this is of course just the bit
     inverse of the destination. */
  /* 0xFFFF - dstdata is obviously just ~dstdata, but try and get the
     compiler to do that without producing a warning, go on... */
  switch (iib->size) {
  case sz_byte:
    XFLAG = CFLAG = srcdata > (0xFFu - (uint8)dstdata);
    break;
  case sz_word:
    XFLAG = CFLAG = srcdata > (0xFFFFu - (uint16)dstdata);
    break;
  case sz_long:
    XFLAG = CFLAG = srcdata > (uint32)~(uint32)dstdata;
    break;
  default:
    break;
  }
  /* gcc doesn't like a one's compliment of an unsigned - take out the
     casts in the above and it will warn you incorrectly */
}

void generate_addxflag_cx(t_iib *iib)
{
  /* V = (Sm && Dm) || (!Rm && (Dm || Sm)) */
  int Sm = 0, Dm = 0, Rm = 0;

  switch(iib->size) {
  case sz_byte:
    Sm = (sint8)srcdata < 0;
    Dm = (sint8)dstdata < 0;
    Rm = (sint8)outdata < 0;
    break;
  case sz_word:
    Sm = (sint16)srcdata < 0;
    Dm = (sint16)dstdata < 0;
    Rm = (sint16)outdata < 0;
    break;
  case sz_long:
    Sm = (sint32)srcdata < 0;
    Dm = (sint32)dstdata < 0;
    Rm = (sint32)outdata < 0;
    break;
  default:
    break;
  }

  XFLAG = CFLAG = (Sm && Dm) || (!Rm && (Dm || Sm));
}

void generate_addflag_v(t_iib *iib)
{
  /* V = (Sm && Dm && !Rm) || (!Sm && !Dm && Rm)
     overflow is performed as if both source and destination are signed,
     the only two condtions are if we've added too much to a +ve number
     or we've subtracted too much from a -ve number */
  /* the technique to implement the above formula is to make sure the sign
     of Sm == the sign of Dm, and the sign of Dm != the sign of Rm. */
  switch(iib->size) {
  case sz_byte:
    VFLAG = (((sint8)srcdata < 0) == ((sint8)dstdata < 0))
         && (((sint8)dstdata < 0) != ((sint8)outdata < 0));
    break;
  case sz_word:
    VFLAG = (((sint16)srcdata < 0) == ((sint16)dstdata < 0))
         && (((sint16)dstdata < 0) != ((sint16)outdata < 0));
    break;
  case sz_long:
    VFLAG = (((sint32)srcdata < 0) == ((sint32)dstdata < 0))
         && (((sint32)dstdata < 0) != ((sint32)outdata < 0));
    break;
  default:
    break;
  }
}

void generate_negflag_cx(t_iib *iib)
{
  /* C = (Dm || Rm) */
  XFLAG = CFLAG = srcdata ? 1 : 0;
}

void generate_negflag_v(t_iib *iib)
{
  /* V = (Dm && Rm)
     which is the same as V = src == 1<<15 as the only case where both the
     data and the result are both negative is in the case of the extreme
     negative number (1<<15 in the case of 16 bit) since 0 - 1<<15 = 1<<15. */
  switch(iib->size) {
  case sz_byte:
    VFLAG = (srcdata == (1u<<7));
    break;
  case sz_word:
    VFLAG = (srcdata == (1u<<15));
    break;
  case sz_long:
    VFLAG = (srcdata == (1u<<31));
    break;
  default:
    break;
  }
}

void generate_negxflag_cx(t_iib *iib)
{
  /* C = (Dm || Rm)
     unlike NEG, NEGX is done properly */
  switch(iib->size) {
  case sz_byte:
    XFLAG = CFLAG = ((sint8)srcdata < 0) || ((sint8)outdata < 0);
    break;
  case sz_word:
    XFLAG = CFLAG = ((sint16)srcdata < 0) || ((sint16)outdata < 0);
    break;
  case sz_long:
    XFLAG = CFLAG = ((sint32)srcdata < 0) || ((sint32)outdata < 0);
    break;
  default:
    break;
  }
}

void generate_negxflag_v(t_iib *iib)
{
  /* V = (Dm && Rm)
     unlike NEG, NEGX is done properly */
  switch(iib->size) {
  case sz_byte:
    VFLAG = ((sint8)srcdata < 0) && ((sint8)outdata < 0);
    break;
  case sz_word:
    VFLAG = ((sint16)srcdata < 0) && ((sint16)outdata < 0);
    break;
  case sz_long:
    VFLAG = ((sint32)srcdata < 0) && ((sint32)outdata < 0);
    break;
  default:
    break;
  }
}

void cctest(void)
{
  uint32 a_cc, b_cc;
  int i, j, pre, err;
  t_iib iib;

  /* cc tests */
  printf("testing cc...\n");
  err = 0;

  for (i=0; i<(1<<7); i++) {
    TFLAG = (i >> 0) & 1;
    SFLAG = (i >> 1) & 1;
    XFLAG = (i >> 2) & 1;
    NFLAG = (i >> 3) & 1;
    ZFLAG = (i >> 4) & 1;
    VFLAG = (i >> 5) & 1;
    CFLAG = (i >> 6) & 1;

    for (j=0; j<16; j++) {
      for (pre=0; pre<2; pre++) {
        if (pre == 0) { /* All flags to zero */
          cc = 0;
        } else {        /* All flags to one */
          cc = 1;
        }
        iib.cc = j;
        generate_cc(&iib);
        a_cc = cc;

        if (pre == 0) { /* All flags to zero */
          cc = 0;
        } else {        /* All flags to one */
          cc = 1;
        }
        iib.cc = j;
        _compile_cc(&iib);
        b_cc = cc;

        if (a_cc != b_cc) {
          printf("cc_error @ %d,%d,%d\n", i, j, pre);
        }
      }
    }
  }


  if (err == 0) {
    printf("\tOK\n");
  } else {
    printf("\tFAILED\n");
  }
}

static int compile_adreg(t_iib *iib, t_ipc *ipc, t_type type)
{
  t_datatype datatype = type ? iib->dtype : iib->stype;

  switch(datatype) {
  case dt_Dreg:
  case dt_Areg:
    return ((datatype == dt_Areg ? 8 : 0) + ((ipc->opcode >> (type ? iib->dbitpos : iib->sbitpos)) & 7));
    break;
  default:
    return 0;
    break;
  }
}

static void compile_ea(t_iib *iib, t_ipc *ipc, t_type type, int update, int *reg, int force)
{
  *reg = -1; /* NO EA */
}

#if 1

static void compile_eaval(t_iib *iib, t_ipc *ipc, t_type type, int eareg, int *eaval, int force)
{
  int reg = compile_adreg(iib, ipc, type);

  OPCODE = lis(3, ((uint32)regsppc >> 16));
  OPCODE = ori(3, 3, ((uint32)regsppc & 0xffff));
  OPCODE = lwz(0, reg*4, 3);

  *eaval = (type ? 12 : 11);

  switch(iib->size) {
  case sz_byte:
    OPCODE = rlwinm(*eaval, 0, 0, 24, 31);
    break;
  case sz_word:
    OPCODE = rlwinm(*eaval, 0, 0, 16, 31);
    break;
  case sz_long:
    OPCODE = rlwinm(*eaval, 0, 0,  0, 31);
    break;
  default:
    printf("EAVAL: Invalid size %d\n", iib->size);
    instr_error = 1;
  }
}

static void compile_eastore(t_iib *iib, t_ipc *ipc, t_type type, int eareg, int eaval)
{
  int reg = compile_adreg(iib, ipc, type);

  OPCODE = lis(3, ((uint32)regsppc >> 16));
  OPCODE = ori(3, 3, ((uint32)regsppc & 0xffff));
  OPCODE = lwz(0, reg*4, 3);

  switch(iib->size) {
  case sz_byte:
    OPCODE = rlwimi(0, eaval, 0, 24, 31);
    break;
  case sz_word:
    OPCODE = rlwimi(0, eaval, 0, 16, 31);
    break;
  case sz_long:
    OPCODE = rlwimi(0, eaval, 0,  0, 31);
    break;
  default:
    printf("EASTORE: Invalid size %d\n", iib->size);
    instr_error = 1;
  }

  OPCODE = stw(0, reg*4, 3);
}

static void compile_adval(t_iib *iib, t_ipc *ipc, t_type type, int *eaval, int force)
{
  int reg = compile_adreg(iib, ipc, type);

  OPCODE = lis(3, ((uint32)regsppc >> 16));
  OPCODE = ori(3, 3, ((uint32)regsppc & 0xffff));
  OPCODE = lwz(0, reg*4, 3);

  *eaval = 10;
  OPCODE = mr(*eaval, 0);
}

static void compile_adstore(t_iib *iib, t_ipc *ipc, t_type type, int eaval)
{
  int reg = compile_adreg(iib, ipc, type);

  OPCODE = mr(0, eaval);

  OPCODE = lis(3, ((uint32)regsppc >> 16));
  OPCODE = ori(3, 3, ((uint32)regsppc & 0xffff));
  OPCODE = stw(0, reg*4, 3);
}

#else

static void compile_eaval(t_iib *iib, t_ipc *ipc, t_type type, int eareg, int *eaval, int force)
{
  int src = (type ? DSTREG : SRCREG);
  *eaval = (type ? 12 : 11);

  switch(iib->size) {
  case sz_byte:
    OPCODE = rlwinm(*eaval, src, 0, 24, 31);
    break;
  case sz_word:
    OPCODE = rlwinm(*eaval, src, 0, 16, 31);
    break;
  case sz_long:
    OPCODE = rlwinm(*eaval, src, 0,  0, 31);
    break;
  default:
    printf("EAVAL: Invalid size %d\n", iib->size);
    instr_error = 1;
  }
}

static void compile_eastore(t_iib *iib, t_ipc *ipc, t_type type, int eareg, int eaval)
{
  int dst = (type ? DSTREG : SRCREG);

  switch(iib->size) {
  case sz_byte:
    OPCODE = rlwimi(dst, eaval, 0, 24, 31);
    break;
  case sz_word:
    OPCODE = rlwimi(dst, eaval, 0, 16, 31);
    break;
  case sz_long:
    OPCODE = rlwimi(dst, eaval, 0,  0, 31);
    break;
  default:
    printf("EASTORE: Invalid size %d\n", iib->size);
    instr_error = 1;
  }
}

static void compile_adval(t_iib *iib, t_ipc *ipc, t_type type, int *eaval, int force)
{
  OPCODE = mr(10, DSTREG);
  *eaval = 10;
}

static void compile_adstore(t_iib *iib, t_ipc *ipc, t_type type, int eaval)
{
  OPCODE = mr(DSTREG, eaval);
}

#endif

/* compiled */

#define DATAREG(a) ((a) * 4)
#define ADDRREG(a) ((8+(a)) * 4)

/* FIXME */
uint32 (*mem68k_fetch_long[0x1000])(uint32 addr);
void (*mem68k_store_long[0x1000])(uint32 addr, uint32 data);

static int _vno;
static uint32 _oldpc;

void reg68k_internal_vector(int vno, uint32 oldpc)
{
  _vno = vno;
  _oldpc = oldpc;
}

static int get_work_reg(int force)
{
  return 12; /* FIXME */
}

static void compile_generic(t_ipc *ipc, t_iib *iib)
{
  int srcreg, dstreg, outreg, srcval, dstval, outval;
  uint32 addr68k = 0;
  void *tmp = NULL, *end = NULL;
  int pc_inc = 1;

  switch(iib->mnemonic) {
/* i_SUB, i_SUBA, i_SUBX */
    #include "compile_op_1.h"
/* i_ADD, i_ADDA, i_ADDX */
    #include "compile_op_2.h"
/* i_ASR, i_LSR, i_ROXR, i_ROR */
    #include "compile_op_3.h"
/* i_ASL, i_LSL, i_ROXL, i_ROL */
    #include "compile_op_4.h"
/* i_BTST, i_BCHG, i_BCLR, i_BSET */
    #include "compile_op_5.h"
/* i_NEG, i_NEGX, i_CLR, i_NOT */
    #include "compile_op_6.h"

/* MISC */
    #include "compile_op_x.h"

    default:
      printf("unknown opcode\n");
      break;
  }
}

/* portable */

void op_store(t_iib *iib, t_type type, uint32 val)
{
  if (type) {
    switch (iib->size) {
    case sz_byte:
       outdata = (dstdata & 0xffffff00) | (val & 0x000000ff);
      _dstdata = (dstdata & 0xffffff00) | (val & 0x000000ff);
      break;
    case sz_word:
       outdata = (dstdata & 0xffff0000) | (val & 0x0000ffff);
      _dstdata = (dstdata & 0xffff0000) | (val & 0x0000ffff);
      break;
    case sz_long:
       outdata = (dstdata & 0x00000000) | (val & 0xffffffff);
      _dstdata = (dstdata & 0x00000000) | (val & 0xffffffff);
      break;
    default:
      break;
    }
  } else {
    switch (iib->size) {
    case sz_byte:
       outdata = (srcdata & 0xffffff00) | (val & 0x000000ff);
      _srcdata = (srcdata & 0xffffff00) | (val & 0x000000ff);
      break;
    case sz_word:
       outdata = (srcdata & 0xffff0000) | (val & 0x0000ffff);
      _srcdata = (srcdata & 0xffff0000) | (val & 0x0000ffff);
      break;
    case sz_long:
       outdata = (srcdata & 0x00000000) | (val & 0xffffffff);
      _srcdata = (srcdata & 0x00000000) | (val & 0xffffffff);
      break;
    default:
      break;
    }
  }
}

void op_divu(t_ipc *ipc, t_iib *iib)
{
#if 0
entry static divu_w_dy_dx
     CYCLES(140,108,42)
     READ_W_REG(r3,rRY)                                // r3 = word-sized rRY
divu_w_ea_common:
     READ_L_REG(r4,rRX)                                // r4 = long-sized rRX
     cmpwi     r3,0                                    // check for divide by zero
     CLR_VC                                                 // clear V & C
     beq       div_exception                           // if zero, generate an exception
     divwu     r5,r4,r3                                // r5 = r4 / r3
     rlwinm.   r0,r5,0,0,15                            // r0 = result & 0xffff
     mullw     r6,r5,r3                                // r6 = r5 * r3
     bne       div_overflow                            // skip if overflow
     cntlzw    r7,r5                                        // count the zeros in the result
     SET_N_WORD(r5)                                    // N = (r5 & 0x8000)
     sub       r6,r4,r6                                // r6 = r4 - r5 * r3
     SET_Z(r7)                                         // Z = (r5 == 0)
     rlwimi    r5,r6,16,0,15                           // pop remainder into high part of the word
     WRITE_L_REG(r5,rRX)                               // long-sized rRX = result
     b         executeLoopEnd                          // all done
div_overflow:
     ori       rSRFlags,rSRFlags,k68000FlagV|k68000FlagN // V = N = 1
     b         executeLoopEnd                          // all done
div_exception:
     li        r3,5                                    // vector 5
     b         generateException                       // generate an exception

      case i_DIVU:
     /* DIVx is the only instruction that has different sizes for the
        source and destination! */
     if (iib->dtype != dt_Dreg)
       OUT("ERROR dtype\n");
     generate_ea(output, iib, tp_src, 1);
     generate_eaval(output, iib, tp_src); /* 16bit EA */
     generate_ea(output, iib, tp_dst, 1); /* 32bit Dn */
        OUT("  uint32 dstdata = DATAREG(dstreg);\n");
        OUT("  uint32 quotient;\n");
     OUT("\n");
     OUT("  if (srcdata == 0) {\n");
     fprintf(output, "    reg68k_internal_vector(V_ZERO, PC+%d);\n",
          (iib->wordlen)*2);
     OUT("    return;\n");
     OUT("  }\n");
     OUT("  quotient = dstdata / srcdata;\n");
        OUT("  if ((quotient & 0xffff0000) == 0) {\n");
     OUT("    DATAREG(dstreg) = quotient | ");
     OUT("(((uint16)(dstdata % srcdata))<<16);\n");
        if (flags && iib->flags.set & IIB_FLAG_V)
          OUT("    VFLAG = 0;\n");
     if (flags && iib->flags.set & IIB_FLAG_N)
       OUT("    NFLAG = ((sint16)quotient) < 0;\n");
     if (flags && iib->flags.set & IIB_FLAG_Z)
       OUT("  ZFLAG = !((uint16)quotient);\n");
     if (flags && (iib->flags.set & IIB_FLAG_V ||
                iib->flags.set & IIB_FLAG_N)) {
       OUT("  } else {\n");
       if (flags && iib->flags.set & IIB_FLAG_V)
         OUT("    VFLAG = 1;\n");
       if (flags && iib->flags.set & IIB_FLAG_N)
         OUT("    NFLAG = 1;\n");
     }
     OUT("  }\n");
     if (flags && iib->flags.set & IIB_FLAG_C)
          OUT("  CFLAG = 0;\n");
     break;
#endif
  uint32 val = 0;

  uint32 quotient;

  if ((uint16)srcdata == 0) {
    reg68k_internal_vector(V_ZERO, PC+(iib->wordlen)*2);
    return;
  }

  quotient = dstdata / (uint16)srcdata;
  if ((quotient & 0xffff0000) == 0) {
    val = quotient | (((uint16)(dstdata % (uint16)srcdata))<<16);

    _dstdata = outdata = val;
  } else { /* overflow */
    if (ipc->set & IIB_FLAG_V)
      VFLAG = 1;
  }
}

void op_divs(t_ipc *ipc, t_iib *iib)
{
#if 0
entry static divs_w_dy_dx
     CYCLES(158,122,54)
     READ_W_REG_EXT(r3,rRY)                            // r3 = word-sized rRY
divs_w_ea_common:
     READ_L_REG(r4,rRX)                                // r4 = long-sized rRX
     cmpwi     r3,0                                    // check for divide by zero
     CLR_VC                                                 // clear V & C
     beq       div_exception                           // if zero, generate an exception
     divw r5,r4,r3                                // r5 = r4 / r3
     rlwinm.   r8,r5,0,0,16                            // r8 = r5 & 0xffff8000
     mullw     r6,r5,r3                                // r6 = r5 * r3
     beq       divs_w_dy_dx_nooverflow                 // skip if no overflow
     not       r8,r5                                        // r8 = ~r5
     rlwinm.   r8,r8,0,0,16                            // r8 = r5 & 0xffff8000
     bne       div_overflow                            // if non-zero this time, it's a real overflow
divs_w_dy_dx_nooverflow:
     sub       r6,r4,r6                                // r6 = r4 - r5 * r3
     cntlzw    r7,r5                                        // count the zeros in the result
     SET_N_WORD(r5)                                    // N = (r5 & 0x8000)
     rlwimi    r5,r6,16,0,15                           // pop remainder into high part of the word
     SET_Z(r7)                                         // Z = (r5 == 0)
     WRITE_L_REG(r5,rRX)                               // long-sized rRX = result
     b         executeLoopEnd                          // all done

      case i_DIVS:
     /* DIVx is the only instruction that has different sizes for the
        source and destination! */
     if (iib->dtype != dt_Dreg)
       OUT("ERROR dtype\n");
     generate_ea(output, iib, tp_src, 1);
     generate_eaval(output, iib, tp_src); /* 16bit EA */
     generate_ea(output, iib, tp_dst, 1); /* 32bit Dn */
     OUT("  sint32 dstdata = DATAREG(dstreg);\n");
     OUT("  sint32 quotient;\n");
     OUT("  sint16 remainder;\n");
     OUT("\n");
     OUT("  if (srcdata == 0) {\n");
     fprintf(output, "    reg68k_internal_vector(V_ZERO, PC+%d);\n",
          (iib->wordlen)*2);
     OUT("    return;\n");
     OUT("  }\n");
     OUT("  quotient = dstdata / (sint16)srcdata;\n");
     OUT("  remainder = dstdata % (sint16)srcdata;\n");
     OUT("  if (((quotient & 0xffff8000) == 0) ||\n");
     OUT("      ((quotient & 0xffff8000) == 0xffff8000)) {\n");
     OUT("    if ((quotient < 0) != (remainder < 0))\n");
     OUT("      remainder = -remainder;\n");
     OUT("    DATAREG(dstreg) = ((uint16)quotient) | ");
     OUT("(((uint16)(remainder))<<16);\n");
        if (flags && iib->flags.set & IIB_FLAG_V)
          OUT("    VFLAG = 0;\n");
     if (flags && iib->flags.set & IIB_FLAG_N)
       OUT("    NFLAG = ((sint16)quotient) < 0;\n");
     if (flags && iib->flags.set & IIB_FLAG_Z)
       OUT("    ZFLAG = !((uint16)quotient);\n");
     if (flags && (iib->flags.set & IIB_FLAG_V ||
                iib->flags.set & IIB_FLAG_N)) {
       OUT("  } else {\n");
       if (flags && iib->flags.set & IIB_FLAG_V)
         OUT("    VFLAG = 1;\n");
       if (flags && iib->flags.set & IIB_FLAG_N)
         OUT("    NFLAG = 1;\n");
     }
     OUT("  }\n");
     if (flags && iib->flags.set & IIB_FLAG_C)
          OUT("  CFLAG = 0;\n");
     break;
#endif
  uint32 val = 0;

  sint32 quotient;
  sint16 remainder;

  if ((sint16)srcdata == 0) {
    reg68k_internal_vector(V_ZERO, PC+(iib->wordlen)*2);
    return;
  }

  quotient = (sint32)dstdata / (sint16)srcdata;
  remainder = (sint32)dstdata % (sint16)srcdata;
  if (((quotient & 0xffff8000) == 0) || ((quotient & 0xffff8000) == 0xffff8000)) {
/* FIXME: BUG in Generator
    if ((quotient < 0) != (remainder < 0)) {
      remainder = -remainder;
    }
*/
    val = ((uint16)quotient) | (((uint16)(remainder))<<16);

    _dstdata = outdata = val;
  } else { /* overflow */
    if (ipc->set & IIB_FLAG_V)
      VFLAG = 1;
  }
}

void op_mulu(t_ipc *ipc, t_iib *iib)
{
/*
      case i_MULU:
     generate_ea(output, iib, tp_src, 1);
     generate_eaval(output, iib, tp_src);
     generate_ea(output, iib, tp_dst, 1);
     generate_eaval(output, iib, tp_dst);
     OUT("\n");
     OUT("  uint32 outdata = (uint32)srcdata * (uint32)dstdata;\n");
     if (iib->dtype != dt_Dreg)
       OUT("ERROR dtype\n");
     OUT("  DATAREG(dstreg) = outdata;\n");
     if (flags)
       OUT("\n");
     if (flags && iib->flags.set & IIB_FLAG_N)
       OUT("  NFLAG = ((sint32)outdata) < 0;\n");
     if (flags && iib->flags.set & IIB_FLAG_Z)
       generate_stdflag_z(output, iib);
     if (flags && iib->flags.set & IIB_FLAG_V)
       generate_clrflag_v(output, iib);
     if (flags && iib->flags.set & IIB_FLAG_C)
       generate_clrflag_c(output, iib);
     break;
*/
  uint32 val = 0;

  val = (uint32)srcdata * (uint32)dstdata;

  iib->size = sz_long;
  op_store(iib, tp_dst, val);

  if (ipc->set & IIB_FLAG_N)
    NFLAG = ((sint32)val) < 0;
  if (ipc->set & IIB_FLAG_Z)
    generate_stdflag_z(iib);
  if (ipc->set & IIB_FLAG_V)
    generate_clrflag_v(iib);
  if (ipc->set & IIB_FLAG_C)
    generate_clrflag_c(iib);
}

void op_muls(t_ipc *ipc, t_iib *iib)
{
/*
      case i_MULS:
     generate_ea(output, iib, tp_src, 1);
     generate_eaval(output, iib, tp_src);
     generate_ea(output, iib, tp_dst, 1);
     generate_eaval(output, iib, tp_dst);
     OUT("\n");
     OUT("  uint32 outdata = (sint32)(sint16)srcdata * "
          "(sint32)(sint16)dstdata;\n");
     if (iib->dtype != dt_Dreg)
       OUT("ERROR dtype\n");
     OUT("  DATAREG(dstreg) = outdata;\n");
     if (flags)
       OUT("\n");
     if (flags && iib->flags.set & IIB_FLAG_N)
       OUT("  NFLAG = ((sint32)outdata) < 0;\n");
     if (flags && iib->flags.set & IIB_FLAG_Z)
       generate_stdflag_z(output, iib);
     if (flags && iib->flags.set & IIB_FLAG_V)
       generate_clrflag_v(output, iib);
     if (flags && iib->flags.set & IIB_FLAG_C)
       generate_clrflag_c(output, iib);
     break;
*/
  uint32 val = 0;

  val = (sint32)(sint16)srcdata * (sint32)(sint16)dstdata;

  iib->size = sz_long;
  op_store(iib, tp_dst, val);

  if (ipc->set & IIB_FLAG_N)
    NFLAG = ((sint32)val) < 0;
  if (ipc->set & IIB_FLAG_Z)
    generate_stdflag_z(iib);
  if (ipc->set & IIB_FLAG_V)
    generate_clrflag_v(iib);
  if (ipc->set & IIB_FLAG_C)
    generate_clrflag_c(iib);
}

void op_cmp(t_ipc *ipc, t_iib *iib)
{
/*
     generate_ea(output, iib, tp_src, 1);
     generate_eaval(output, iib, tp_src);
     generate_ea(output, iib, tp_dst, 1);
     generate_eaval(output, iib, tp_dst);
     switch (iib->size) {
     case sz_byte:
       generate_outdata(output, iib, "(sint8)dstdata - (sint8)srcdata");
       break;
     case sz_word:
       generate_outdata(output, iib, "(sint16)dstdata - (sint16)srcdata");
       break;
     case sz_long:
       generate_outdata(output, iib, "(sint32)dstdata - (sint32)srcdata");
       break;
     default:
       OUT("ERROR size\n");
       break;
     }
     if (flags)
       OUT("\n");
     if (flags && iib->flags.set & IIB_FLAG_V)
       generate_subflag_v(output, iib);
     if (flags && iib->flags.set & IIB_FLAG_C)
       generate_subflag_c(output, iib);
     if (flags && iib->flags.set & IIB_FLAG_N)
       generate_stdflag_n(output, iib);
     if (flags && iib->flags.set & IIB_FLAG_Z)
       generate_stdflag_z(output, iib);
*/
  uint32 val = 0, outbak = outdata, _dstbak = _dstdata;

  switch (iib->size) {
  case sz_byte:
    val = (sint8)dstdata - (sint8)srcdata;
    break;
  case sz_word:
    val = (sint16)dstdata - (sint16)srcdata;
    break;
  case sz_long:
    val = (sint32)dstdata - (sint32)srcdata;
    break;
  default:
    break;
  }

  op_store(iib, tp_dst, val); /* FIXME: wrong */

  if (ipc->set & IIB_FLAG_V)
    generate_subflag_v(iib);
  if (ipc->set & IIB_FLAG_C)
    generate_subflag_c(iib);
  if (ipc->set & IIB_FLAG_N)
    generate_stdflag_n(iib);
  if (ipc->set & IIB_FLAG_Z)
    generate_stdflag_z(iib);

   outdata =  outbak; /* FIXME: hack due to op_store above */
  _dstdata = _dstbak;
}

void op_cmpa(t_ipc *ipc, t_iib *iib)
{
/*
     if (iib->dtype != dt_Areg || iib->size != sz_word)
       OUT("Error\n");
     generate_ea(output, iib, tp_src, 1);
     generate_eaval(output, iib, tp_src);
     iib->size = sz_long;
     generate_ea(output, iib, tp_dst, 1);
     generate_eaval(output, iib, tp_dst);
        OUT("  uint32 outdata = (sint32)dstdata - (sint32)(sint16)srcdata;\n");
     if (flags)
       OUT("\n");
     if (flags && iib->flags.set & IIB_FLAG_V)
       generate_cmpaflag_v(output, iib);
     if (flags && iib->flags.set & IIB_FLAG_C)
       generate_cmpaflag_c(output, iib);
     if (flags && iib->flags.set & IIB_FLAG_N)
       generate_stdflag_n(output, iib);
     if (flags && iib->flags.set & IIB_FLAG_Z)
       generate_stdflag_z(output, iib);
     iib->size = sz_word;
*/

  uint32 val = 0, outbak = outdata, _dstbak = _dstdata;

  val = (sint32)dstdata - (sint32)(sint16)srcdata;

  op_store(iib, tp_dst, val); /* FIXME: wrong */

  if (ipc->set & IIB_FLAG_V)
    generate_cmpaflag_v(iib);
  if (ipc->set & IIB_FLAG_C)
    generate_cmpaflag_c(iib);
  if (ipc->set & IIB_FLAG_N)
    generate_stdflag_n(iib);
  if (ipc->set & IIB_FLAG_Z)
    generate_stdflag_z(iib);

   outdata =  outbak; /* FIXME: hack due to op_store above */
  _dstdata = _dstbak;
}

void op_negx(t_ipc *ipc, t_iib *iib)
{
  uint32 val = 0;

  switch (iib->size) {
  case sz_byte:
    val = 0 - (sint8)srcdata - XFLAG;
    break;
  case sz_word:
    val = 0 - (sint16)srcdata - XFLAG;
    break;
  case sz_long:
    val = 0 - (sint32)srcdata - XFLAG;
    break;
  default:
    break;
  }

  op_store(iib, tp_src, val);

  if (ipc->set & IIB_FLAG_V)
    generate_negxflag_v(iib);
  if (((ipc->set & IIB_FLAG_C) || (ipc->set & IIB_FLAG_X)))
    generate_negxflag_cx(iib);
  if (ipc->set & IIB_FLAG_N)
    generate_stdflag_n(iib);
  if (ipc->set & IIB_FLAG_Z)
    generate_stdxflag_z(iib);
}

void op_clr(t_ipc *ipc, t_iib *iib)
{
  op_store(iib, tp_src, 0);

  if (ipc->set & IIB_FLAG_V)
    generate_clrflag_v(iib);
  if (ipc->set & IIB_FLAG_C)
    generate_clrflag_c(iib);
  if (ipc->set & IIB_FLAG_N)
    generate_clrflag_n(iib);
  if (ipc->set & IIB_FLAG_Z)
    generate_setflag_z(iib);
}

void op_bxxx(t_ipc *ipc, t_iib *iib)
{
  uint32 bitpos = 0;
  uint32 val;

  switch (iib->size) {
  case sz_byte:
    bitpos = 1<<(srcdata & 7);
    break;
  case sz_long:
    bitpos = 1<<(srcdata & 31);
    break;
  default:
     break;
   }

  switch(iib->mnemonic) {
  case i_BTST:
    break;
  case i_BCHG:
    val = dstdata ^ bitpos;
    op_store(iib, tp_dst, val);
    break;
  case i_BCLR:
    val = dstdata & ~bitpos;
    op_store(iib, tp_dst, val);
    break;
  case i_BSET:
    val = dstdata | bitpos;
    op_store(iib, tp_dst, val);
    break;
  default:
    break;
  }

  if (ipc->set & IIB_FLAG_Z)
    ZFLAG = !(dstdata & bitpos);
}

void op_roxl(t_ipc *ipc, t_iib *iib)
{
  uint8 bits = (iib->size == sz_byte) ? 8 : (iib->size == sz_word) ? 16 : 32;
  uint8 loop = srcdata & 63;
  uint8 cflag = CFLAG;
  uint8 xflag = XFLAG;
  uint32 val = dstdata;

  while(loop) {
    cflag = val & 1<<(bits-1) ? 1 : 0;
    val<<= 1;
    val |= xflag;
    xflag = cflag;
    loop--;
  }

  op_store(iib, tp_dst, val);

  if (ipc->set & IIB_FLAG_X)
    XFLAG = xflag;
  if (ipc->set & IIB_FLAG_C)
    CFLAG = xflag;
  if (ipc->set & IIB_FLAG_N)
    generate_stdflag_n(iib);
  if (ipc->set & IIB_FLAG_Z)
    generate_stdflag_z(iib);
  if (ipc->set & IIB_FLAG_V)
    generate_clrflag_v(iib);
}

void op_roxr(t_ipc *ipc, t_iib *iib)
{
  uint8 bits = (iib->size == sz_byte) ? 8 : (iib->size == sz_word) ? 16 : 32;
  uint8 loop = srcdata & 63;
  uint8 cflag = CFLAG;
  uint8 xflag = XFLAG;
  uint32 val = dstdata;

  while(loop) {
    cflag = val & 1;
    val>>= 1;
    if (xflag)
      val |= 1<<(bits-1);
    xflag = cflag;
    loop--;
  }

  op_store(iib, tp_dst, val);

  if (ipc->set & IIB_FLAG_X)
    XFLAG = xflag;
  if (ipc->set & IIB_FLAG_C)
    CFLAG = xflag;
  if (ipc->set & IIB_FLAG_N)
    generate_stdflag_n(iib);
  if (ipc->set & IIB_FLAG_Z)
    generate_stdflag_z(iib);
  if (ipc->set & IIB_FLAG_V)
    generate_clrflag_v(iib);
}

void op_rol(t_ipc *ipc, t_iib *iib)
{
  uint8 bits = (iib->size == sz_byte) ? 8 : (iib->size == sz_word) ? 16 : 32;
  uint8 loop = srcdata & 63;
  uint8 cflag = 0;
  uint32 val = dstdata;

  while(loop) {
    cflag = val & 1<<(bits-1) ? 1 : 0;
    val<<= 1;
    if (cflag)
      val |= 1;
    loop--;
  }

  op_store(iib, tp_dst, val);

  if (ipc->set & IIB_FLAG_C)
    CFLAG = cflag;
  if (ipc->set & IIB_FLAG_N)
    generate_stdflag_n(iib);
  if (ipc->set & IIB_FLAG_Z)
    generate_stdflag_z(iib);
  if (ipc->set & IIB_FLAG_V)
    generate_clrflag_v(iib);
}

void op_ror(t_ipc *ipc, t_iib *iib)
{
  uint8 bits = (iib->size == sz_byte) ? 8 : (iib->size == sz_word) ? 16 : 32;
  uint8 loop = srcdata & 63;
  uint8 cflag = 0;
  uint32 val = dstdata;

  while(loop) {
    cflag = val & 1;
    val>>= 1;
    if (cflag)
      val |= 1<<(bits-1);
    loop--;
  }

  op_store(iib, tp_dst, val);

  if (ipc->set & IIB_FLAG_C)
    CFLAG = cflag;
  if (ipc->set & IIB_FLAG_N)
    generate_stdflag_n(iib);
  if (ipc->set & IIB_FLAG_Z)
    generate_stdflag_z(iib);
  if (ipc->set & IIB_FLAG_V)
    generate_clrflag_v(iib);
}

void op_lsl(t_ipc *ipc, t_iib *iib)
{
  uint8 bits = (iib->size == sz_byte) ? 8 : (iib->size == sz_word) ? 16 : 32;
  uint8 count = srcdata & 63;
  uint32 val = 0;

  val = count >= bits ? 0 : (dstdata << count);

  op_store(iib, tp_dst, val);

  if (((ipc->set & IIB_FLAG_C) || (ipc->set & IIB_FLAG_X))) {

  if (!count)
    CFLAG = 0;
  else if (count >= bits) {
    CFLAG = (count == bits) ? dstdata & 1 : 0;
    XFLAG = (count == bits) ? dstdata & 1 : 0;
  } else {
    CFLAG = dstdata>>(bits-count) & 1;
    XFLAG = dstdata>>(bits-count) & 1;
  }

  }

  if (ipc->set & IIB_FLAG_V)
    generate_clrflag_v(iib);
  if (ipc->set & IIB_FLAG_N)
    generate_stdflag_n(iib);
  if (ipc->set & IIB_FLAG_Z)
    generate_stdflag_z(iib);
}

void op_lsr(t_ipc *ipc, t_iib *iib)
{
  uint8 bits = (iib->size == sz_byte) ? 8 : (iib->size == sz_word) ? 16 : 32;
  uint8 count = srcdata & 63;
  uint32 val = 0;

//  val = dstdata >> (count > (bits-1) ? (bits-1) : count); // BUG
  val = dstdata >> count;

  op_store(iib, tp_dst, val);

  if (((ipc->set & IIB_FLAG_C) || (ipc->set & IIB_FLAG_X))) {

  if (!count)
    CFLAG = 0;
  else if (count >= bits) {
    CFLAG = (count == bits) ? dstdata>>(bits-1) : 0;
    XFLAG = (count == bits) ? dstdata>>(bits-1) : 0;
  } else {
    CFLAG = dstdata>>(count-1) & 1;
    XFLAG = dstdata>>(count-1) & 1;
  }

  }

  if (ipc->set & IIB_FLAG_V)
    generate_clrflag_v(iib);
  if (ipc->set & IIB_FLAG_N)
    generate_stdflag_n(iib);
  if (ipc->set & IIB_FLAG_Z)
    generate_stdflag_z(iib);
}

void op_asl(t_ipc *ipc, t_iib *iib)
{
  uint8 bits = (iib->size == sz_byte) ? 8 : (iib->size == sz_word) ? 16 : 32;
  uint8 count = srcdata & 63;
  uint32 val = 0;

  val = count >= bits ? 0 : (dstdata << count);

  op_store(iib, tp_dst, val);

  if (((ipc->set & IIB_FLAG_C) || (ipc->set & IIB_FLAG_X))) {

  if (!count) {
    CFLAG = 0;
VFLAG = 0; // BUG in generator? Should be this way according to docs...
  } else if (count >= bits) {
    CFLAG = (count == bits) ? dstdata & 1 : 0;
    XFLAG = (count == bits) ? dstdata & 1 : 0;
//    VFLAG = !dstdata; // BUG ???
    VFLAG = dstdata != 0;
  } else {
    CFLAG = dstdata>>(bits-count) & 1;
    XFLAG = dstdata>>(bits-count) & 1;
    switch (iib->size) {
    case sz_byte: {
      uint8 mask = 0xff << (7-count);
      VFLAG = ((dstdata & mask) != mask) && ((dstdata & mask) != 0);
      } break;
    case sz_word: {
      uint16 mask = 0xffff << (15-count);
      VFLAG = ((dstdata & mask) != mask) && ((dstdata & mask) != 0);
      } break;
    case sz_long: {
      uint32 mask = 0xffffffff << (31-count);
      VFLAG = ((dstdata & mask) != mask) && ((dstdata & mask) != 0);
      } break;
    default:
      break;
    }
  }

  }

  if (ipc->set & IIB_FLAG_N)
    generate_stdflag_n(iib);
  if (ipc->set & IIB_FLAG_Z)
    generate_stdflag_z(iib);
}

void op_asr(t_ipc *ipc, t_iib *iib)
{
//  uint8 bits = (iib->size == sz_byte) ? 8 : (iib->size == sz_word) ? 16 : 32;
  uint8 count = srcdata & 63;
  uint32 val = 0;

  switch (iib->size) {
  case sz_byte:
    val = ((sint8)dstdata) >> (count > 7 ? 7 : count);
    break;
  case sz_word:
    val = ((sint16)dstdata) >> (count > 15 ? 15 : count);
    break;
  case sz_long:
    val = ((sint32)dstdata) >> (count > 31 ? 31 : count);
    break;
  default:
    break;
  }

  op_store(iib, tp_dst, val);

  if (((ipc->set & IIB_FLAG_C) || (ipc->set & IIB_FLAG_X))) {

#if 1 // new bugs found...

  if (!count) {
    CFLAG = 0;
//  } else if (count >= bits) { // BUG
//    CFLAG = dstdata>>(bits-1);
//    XFLAG = dstdata>>(bits-1);
  } else {
    count--;

    switch (iib->size) {
    case sz_byte:
      CFLAG = XFLAG = (((sint8)dstdata) >> (count > 7 ? 7 : count)) & 1;
      break;
    case sz_word:
      CFLAG = XFLAG = (((sint16)dstdata) >> (count > 15 ? 15 : count)) & 1;
      break;
    case sz_long:
      CFLAG = XFLAG = (((sint32)dstdata) >> (count > 31 ? 31 : count)) & 1;
      break;
    default:
      break;
    }
  }

#else
#if 0 // I'm pretty sure this is a CPU emulation bug in Generator...

  if (!srcdata) // BUG
    CFLAG = 0;
//  else if (srcdata >= bits) {
//    CFLAG = dstdata>>(bits-1);
//    XFLAG = dstdata>>(bits-1);
//  } else {
  else {
    CFLAG = dstdata>>(count-1) & 1;
    XFLAG = dstdata>>(count-1) & 1;
  }

#else

  if (!count) {
    CFLAG = 0;
//  } else if (count >= bits) { // BUG
//    CFLAG = dstdata>>(bits-1);
//    XFLAG = dstdata>>(bits-1);
  } else {
    CFLAG = dstdata>>(count-1) & 1;
    XFLAG = dstdata>>(count-1) & 1;
  }

#endif
#endif
  }

  if (ipc->set & IIB_FLAG_V)
    generate_clrflag_v(iib);
  if (ipc->set & IIB_FLAG_N)
    generate_stdflag_n(iib);
  if (ipc->set & IIB_FLAG_Z)
    generate_stdflag_z(iib);
}

void op_neg(t_ipc *ipc, t_iib *iib)
{
  uint32 val = 0;

  switch (iib->size) {
  case sz_byte:
    val = 0 - (sint8)srcdata;
    break;
  case sz_word:
    val = 0 - (sint16)srcdata;
    break;
  case sz_long:
    val = 0 - (sint32)srcdata;
    break;
  default:
    break;
  }

  op_store(iib, tp_src, val);

  if (ipc->set & IIB_FLAG_V)
    generate_negflag_v(iib);
  if (((ipc->set & IIB_FLAG_C) || (ipc->set & IIB_FLAG_X)))
    generate_negflag_cx(iib);
  if (ipc->set & IIB_FLAG_N)
    generate_stdflag_n(iib);
  if (ipc->set & IIB_FLAG_Z)
    generate_stdflag_z(iib);
}

void op_not(t_ipc *ipc, t_iib *iib)
{
  uint32 val = 0;

  val = ~srcdata;

  op_store(iib, tp_src, val);

  if (ipc->set & IIB_FLAG_V)
    generate_clrflag_v(iib);
  if (ipc->set & IIB_FLAG_C)
    generate_clrflag_c(iib);
  if (ipc->set & IIB_FLAG_N)
    generate_stdflag_n(iib);
  if (ipc->set & IIB_FLAG_Z)
    generate_stdflag_z(iib);
}

void op_subx(t_ipc *ipc, t_iib *iib)
{
  uint32 val = 0;

  switch (iib->size) {
  case sz_byte:
    val = (sint8)dstdata - (sint8)srcdata - XFLAG;
    break;
  case sz_word:
    val = (sint16)dstdata - (sint16)srcdata - XFLAG;
    break;
  case sz_long:
    val = (sint32)dstdata - (sint32)srcdata - XFLAG;
    break;
  default:
    break;
  }

  op_store(iib, tp_dst, val);

  if (ipc->set & IIB_FLAG_V)
    generate_subflag_v(iib);
  if (((ipc->set & IIB_FLAG_C) || (ipc->set & IIB_FLAG_X)))
    generate_subxflag_cx(iib);
  if (ipc->set & IIB_FLAG_N)
    generate_stdflag_n(iib);
  if (ipc->set & IIB_FLAG_Z)
    generate_stdxflag_z(iib);
}

void op_suba(t_ipc *ipc, t_iib *iib)
{
  switch (iib->size) {
  case sz_byte:
    _dstdata = outdata = (sint32)dstdata - (sint8)srcdata;
    break;
  case sz_word:
    _dstdata = outdata = (sint32)dstdata - (sint16)srcdata;
    break;
  case sz_long:
    _dstdata = outdata = (sint32)dstdata - (sint32)srcdata;
    break;
  default:
    break;
  }
}

void op_sub(t_ipc *ipc, t_iib *iib)
{
  uint32 val = 0;

  switch (iib->size) {
  case sz_byte:
    val = (sint8)dstdata - (sint8)srcdata;
    break;
  case sz_word:
    val = (sint16)dstdata - (sint16)srcdata;
    break;
  case sz_long:
    val = (sint32)dstdata - (sint32)srcdata;
    break;
  default:
    break;
  }

  op_store(iib, tp_dst, val);

  if (ipc->set & IIB_FLAG_V)
    generate_subflag_v(iib);
  if (((ipc->set & IIB_FLAG_C) || (ipc->set & IIB_FLAG_X)))
    generate_subflag_cx(iib);
  if (ipc->set & IIB_FLAG_N)
    generate_stdflag_n(iib);
  if (ipc->set & IIB_FLAG_Z)
    generate_stdflag_z(iib);
}

void op_orandeor(t_ipc *ipc, t_iib *iib)
{
  uint32 val = 0;

  switch(iib->mnemonic) {
  case i_OR:
    val = dstdata | srcdata;
    break;
  case i_AND:
    val = dstdata & srcdata;
    break;
  case i_EOR:
    val = dstdata ^ srcdata;
    break;
  default:
    break;
  }

  op_store(iib, tp_dst, val);

  if (ipc->set & IIB_FLAG_N)
    generate_stdflag_n(iib);
  if (ipc->set & IIB_FLAG_Z)
    generate_stdflag_z(iib);
  if (ipc->set & IIB_FLAG_V)
    generate_clrflag_v(iib);
  if (ipc->set & IIB_FLAG_C)
    generate_clrflag_c(iib);
}

void op_addx(t_ipc *ipc, t_iib *iib)
{
  uint32 val = 0;

  switch (iib->size) {
  case sz_byte:
    val = (sint8)dstdata + (sint8)srcdata + XFLAG;
    break;
  case sz_word:
    val = (sint16)dstdata + (sint16)srcdata + XFLAG;
    break;
  case sz_long:
    val = (sint32)dstdata + (sint32)srcdata + XFLAG;
    break;
  default:
    break;
  }

  op_store(iib, tp_dst, val);

  if (ipc->set & IIB_FLAG_V)
    generate_addflag_v(iib);
  if (((ipc->set & IIB_FLAG_C) || (ipc->set & IIB_FLAG_X)))
    generate_addxflag_cx(iib);
  if (ipc->set & IIB_FLAG_N)
    generate_stdflag_n(iib);
  if (ipc->set & IIB_FLAG_Z)
    generate_stdxflag_z(iib);
}

void op_adda(t_ipc *ipc, t_iib *iib)
{
  switch (iib->size) {
  case sz_byte:
    _dstdata = outdata = (sint32)dstdata + (sint8)srcdata;
    break;
  case sz_word:
    _dstdata = outdata = (sint32)dstdata + (sint16)srcdata;
    break;
  case sz_long:
    _dstdata = outdata = (sint32)dstdata + (sint32)srcdata;
    break;
  default:
    break;
  }
}

void op_add(t_ipc *ipc, t_iib *iib)
{
  uint32 val = 0;

  switch (iib->size) {
  case sz_byte:
    val = (sint8)dstdata + (sint8)srcdata;
    break;
  case sz_word:
    val = (sint16)dstdata + (sint16)srcdata;
    break;
  case sz_long:
    val = (sint32)dstdata + (sint32)srcdata;
    break;
  default:
    break;
  }

  op_store(iib, tp_dst, val);

  if (ipc->set & IIB_FLAG_V)
    generate_addflag_v(iib);
  if (((ipc->set & IIB_FLAG_C) || (ipc->set & IIB_FLAG_X)))
    generate_addflag_cx(iib);
  if (ipc->set & IIB_FLAG_N)
    generate_stdflag_n(iib);
  if (ipc->set & IIB_FLAG_Z)
    generate_stdflag_z(iib);
}

/* optest */

void compile_op_init(void)
{
  memset(buf, 0, BUFSIZE);
  ptr = buf;

  OPCODE = stwu(1, -240, 1);
  OPCODE = mflr(0);
  OPCODE = stw(0, 244, 1);

  OPCODE = stw(28, 24 + ((28 - 14) * 4), 1);
  OPCODE = stw(29, 24 + ((29 - 14) * 4), 1);

  OPCODE = lis(3, ((uint32)&srcdata >> 16));
  OPCODE = ori(3, 3, ((uint32)&srcdata & 0xffff));
  OPCODE = lwz(SRCREG, 0, 3);

  OPCODE = lis(3, ((uint32)&dstdata >> 16));
  OPCODE = ori(3, 3, ((uint32)&dstdata & 0xffff));
  OPCODE = lwz(DSTREG, 0, 3);
}

void compile_op_exit(void)
{
  OPCODE = lis(3, ((uint32)&srcdata >> 16));
  OPCODE = ori(3, 3, ((uint32)&srcdata & 0xffff));
  OPCODE = stw(SRCREG, 0, 3);

  OPCODE = lis(3, ((uint32)&dstdata >> 16));
  OPCODE = ori(3, 3, ((uint32)&dstdata & 0xffff));
  OPCODE = stw(DSTREG, 0, 3);

  OPCODE = lwz(28, 24 + ((28 - 14) * 4), 1);
  OPCODE = lwz(29, 24 + ((29 - 14) * 4), 1);

  OPCODE = lwz(0, 244, 1);
  OPCODE = mtlr(0);
  OPCODE = la(1, 240, 1);
  OPCODE = blr();

  /* Flush the cache */
  SetCache(CACHE_ICACHEINV, 0, 0);
  SetCache(CACHE_DCACHEFLUSH, 0, 0);
}

struct optest_s {
  uint8 *name;
  void (*reference)(t_ipc *ipc, t_iib *iib);
  t_mnemonic mnemonic;
  int size[3];
} optests[] = {
  { "sub", op_sub, i_SUB, { sz_byte, sz_word, sz_long } },
  { "suba", op_suba, i_SUBA, { sz_word, sz_long, sz_none } },
  { "subx", op_subx, i_SUBX, { sz_byte, sz_word, sz_long } },

  { "add", op_add, i_ADD, { sz_byte, sz_word, sz_long } },
  { "adda", op_adda, i_ADDA, { sz_word, sz_long, sz_none } },
  { "addx", op_addx, i_ADDX, { sz_byte, sz_word, sz_long } },

  { "asr", op_asr, i_ASR, { sz_byte, sz_word, sz_long } },
  { "lsr", op_lsr, i_LSR, { sz_byte, sz_word, sz_long } },
  { "roxr", op_roxr, i_ROXR, { sz_byte, sz_word, sz_long } },
  { "ror", op_ror, i_ROR, { sz_byte, sz_word, sz_long } },

  { "asl", op_asl, i_ASL, { sz_byte, sz_word, sz_long } },
  { "lsl", op_lsl, i_LSL, { sz_byte, sz_word, sz_long } },
  { "roxl", op_roxl, i_ROXL, { sz_byte, sz_word, sz_long } },
  { "rol", op_rol, i_ROL, { sz_byte, sz_word, sz_long } },

  { "or", op_orandeor, i_OR, { sz_byte, sz_word, sz_long } },
  { "and", op_orandeor, i_AND, { sz_byte, sz_word, sz_long } },
  { "eor", op_orandeor, i_EOR, { sz_byte, sz_word, sz_long } },

  { "cmp", op_cmp, i_CMP, { sz_byte, sz_word, sz_long } },
  { "cmpa", op_cmpa, i_CMPA, { sz_word, sz_none, sz_none } },

  { "neg", op_neg, i_NEG, { sz_byte, sz_word, sz_long } },
  { "negx", op_negx, i_NEGX, { sz_byte, sz_word, sz_long } },
  { "clr", op_clr, i_CLR, { sz_byte, sz_word, sz_long } },
  { "not", op_not, i_NOT, { sz_byte, sz_word, sz_long } },

  { "btst", op_bxxx, i_BTST, { sz_byte, sz_long, sz_none } },
  { "bchg", op_bxxx, i_BCHG, { sz_byte, sz_long, sz_none } },
  { "bclr", op_bxxx, i_BCLR, { sz_byte, sz_long, sz_none } },
  { "bset", op_bxxx, i_BSET, { sz_byte, sz_long, sz_none } },

  { "mulu", op_mulu, i_MULU, { sz_word, sz_none, sz_none } },
  { "muls", op_muls, i_MULS, { sz_word, sz_none, sz_none } },

  { "divu", op_divu, i_DIVU, { sz_word, sz_none, sz_none } }, /* 32 bits / 16 bits */
  { "divs", op_divs, i_DIVS, { sz_word, sz_none, sz_none } }, /* 32 bits / 16 bits */

/* END */
  { NULL, NULL, NULL },
};

static char cflags[2][8];
static int cflags_num = 0;
static char *showflags(t_sr sr)
{
  char *ptr;

  cflags_num ^= 1;
  ptr = cflags[cflags_num];

  if (sr.sr_struct.t) ptr[0] = 'T'; else ptr[0] = '-';
  if (sr.sr_struct.s) ptr[1] = 'S'; else ptr[1] = '-';
  if (sr.sr_struct.x) ptr[2] = 'X'; else ptr[2] = '-';
  if (sr.sr_struct.n) ptr[3] = 'N'; else ptr[3] = '-';
  if (sr.sr_struct.z) ptr[4] = 'Z'; else ptr[4] = '-';
  if (sr.sr_struct.v) ptr[5] = 'V'; else ptr[5] = '-';
  if (sr.sr_struct.c) ptr[6] = 'C'; else ptr[6] = '-';
  ptr[7] = '\0';

  return ptr;
}

static void build68k(unsigned short *ptr, unsigned short op)
{
  *ptr++ = 0x44C7; // move d7,ccr

  *ptr++ = op;     // op

  *ptr++ = 0x42C7; // move ccr,d7

  *ptr++ = 0x4E75; // rts

  CacheClearU(); /* flush any cache */
}

/* e.x. add.b d0,d1 (d0 = src, d1 = dst), adda.w d0,a0 (d0 = src, a0 = dst) */

static uint16 ops[] = {

#if 1

/* D */

0x9200,//                          51      sub.b d0,d1
0x9240,//                          52      sub.w d0,d1
0x9280,//                          53      sub.l d0,d1
0x9300,//                          59      subx.b d0,d1
0x9340,//                          60      subx.w d0,d1
0x9380,//                          61      subx.l d0,d1
0xD200,//                          63      add.b d0,d1
0xD240,//                          64      add.w d0,d1
0xD280,//                          65      add.l d0,d1
0xD300,//                          71      addx.b d0,d1
0xD340,//                          72      addx.w d0,d1
0xD380,//                          73      addx.l d0,d1
0xE021,//                          75      asr.b d0,d1
0xE029,//                          76      lsr.b d0,d1
0xE031,//                          77      roxr.b d0,d1
0xE039,//                          78      ror.b d0,d1
0xE061,//                          80      asr.w d0,d1
0xE069,//                          81      lsr.w d0,d1
0xE071,//                          82      roxr.w d0,d1
0xE079,//                          83      ror.w d0,d1
0xE0A1,//                          85      asr.l d0,d1
0xE0A9,//                          86      lsr.l d0,d1
0xE0B1,//                          87      roxr.l d0,d1
0xE0B9,//                          88      ror.l d0,d1
0xE121,//                          90      asl.b d0,d1
0xE129,//                          91      lsl.b d0,d1
0xE131,//                          92      roxl.b d0,d1
0xE139,//                          93      rol.b d0,d1
0xE161,//                          95      asl.w d0,d1
0xE169,//                          96      lsl.w d0,d1
0xE171,//                          97      roxl.w d0,d1
0xE179,//                          98      rol.w d0,d1
0xE1A1,//                         100      asl.l d0,d1
0xE1A9,//                         101      lsl.l d0,d1
0xE1B1,//                         102      roxl.l d0,d1
0xE1B9,//                         103      rol.l d0,d1
0x8200,//                         105      or.b d0,d1
0xC200,//                         106      and.b d0,d1
0xB101,//                         107      eor.b d0,d1
0x8240,//                         109      or.w d0,d1
0xC240,//                         110      and.w d0,d1
0xB141,//                         111      eor.w d0,d1
0x8280,//                         113      or.l d0,d1
0xC280,//                         114      and.l d0,d1
0xB181,//                         115      eor.l d0,d1
0xB200,//                         117      cmp.b d0,d1
0xB240,//                         120      cmp.w d0,d1
0xB280,//                         123      cmp.l d0,d1

0x4400,//                         126      neg.b d0
0x4000,//                         127      negx.b d0
0x4200,//                         128      clr.b d0
0x4600,//                         129      not.b d0
0x4440,//                         131      neg.w d0
0x4040,//                         132      negx.w d0
0x4240,//                         133      clr.w d0
0x4640,//                         134      not.w d0
0x4480,//                         136      neg.l d0
0x4080,//                         137      negx.l d0
0x4280,//                         138      clr.l d0
0x4680,//                         139      not.l d0

0x0101,//                         141      btst.b d0,d1
0x0141,//                         142      bchg.b d0,d1
0x0181,//                         143      bclr.b d0,d1
0x01C1,//                         144      bset.b d0,d1
0x0101,//                         146      btst.w d0,d1
0x0141,//                         147      bchg.w d0,d1
0x0181,//                         148      bclr.w d0,d1
0x01C1,//                         149      bset.w d0,d1
0x0101,//                         151      btst.l d0,d1
0x0141,//                         152      bchg.l d0,d1
0x0181,//                         153      bclr.l d0,d1
0x01C1,//                         154      bset.l d0,d1
0xC2C0,//                         156      mulu.w d0,d1
0xC3C0,//                         157      muls.w d0,d1
0x82C0,//                         168      divu.w d0,d1
0x83C0,//                         169      divs.w d0,d1
0x1200,//                         174      move.b d0,d1
0x3200,//                         175      move.w d0,d1
0x2200,//                         176      move.l d0,d1

/* A */

0x90C0,//                          56      suba.w d0,a0
0x91C0,//                          57      suba.l d0,a0
0xD0C0,//                          68      adda.w d0,a0
0xD1C0,//                          69      adda.l d0,a0
0xB0C0,//                         121      cmpa.w d0,a0
0xB1C0,//                         124      cmpa.l d0,a0
0x3040,//                         179      movea.w d0,a0
0x2040,//                         180      movea.l d0,a0

#endif

/* NEW */

0x4840,//                         183      swap.w d0
0x4880,//                         187      ext.w d0
0x48C0,//                         188      ext.l d0
0xC141,//                         192      exg.l d0,d1
0xC149,//                         196      exg.l a0,a1
0xC189,//                         200      exg.l d0,a1
0xC388,//                         204      exg.l a0,d1
0x4A00,//                         206      tst.b d0
0x4A40,//                         207      tst.w d0
0x4A80,//                         208      tst.l d0
0x4AC0,//                         210      tas.b d0

/* END */
  0,
};

void optest_real68k(void)
{
  int i, j, src, dst, pre, set, err;
  uint32 *src_data, *dst_data;
  t_sr a_sr, b_sr, set_sr, clr_sr, sr_used, sr_set;

  static struct PPCArgs PP;
  static unsigned short buf68k[1024];
  static char diss_68k[1024];

  t_ipc ipc;
  t_iib *iib;
  op_t op;

  cpu68k_init();

  /* optests */
  i = 0;
  while (ops[i] != 0) {
    iib = cpu68k_iibtable[ops[i]];
    if (iib == NULL) {
      printf("testing invalid instruction\n");
      continue;
    } else {
      cpu68k_ipc(0, (uint8 *)&ops[i], iib, &ipc);
      diss68k_gettext(&ipc, diss_68k);
      sr_used.sr_int = iib->flags.used & 0x1f;
      sr_set.sr_int = iib->flags.set & 0x1f;
      printf("testing %s (used: %s, set: %s)\n", diss_68k, showflags(sr_used), showflags(sr_set));
    }

    build68k(buf68k, ops[i]);

    compile_op_init();
    compile_generic(&ipc, iib);
    compile_op_exit();
    op = (op_t)buf;

    err = 0;
    set_sr.sr_int = 0;
    clr_sr.sr_int = 0;

    switch (iib->mnemonic) {
      case i_DIVS:
      case i_DIVU:
        src_data = word_data;
        dst_data = long_data;
        break;

      default:
        src_data = (iib->size == sz_byte) ? byte_data : (iib->size == sz_word) ? word_data : long_data;
        dst_data = (iib->size == sz_byte) ? byte_data : (iib->size == sz_word) ? word_data : long_data;
        break;
    }

    for (src=0; src<DATAS; src++) {
      if (src_data[src] == 0) {
        switch (iib->mnemonic) {
          case i_DIVS:
          case i_DIVU:
            continue; /* we don't want any exceptions */

          default:
            break;
        }
      }

      for (dst=0; dst<DATAS; dst++) {
        for (pre=0; pre<2; pre++) {
          for (set=0; set<=31; set+=31) {
            ipc.set = set;

            memset(regs68k, 0, sizeof(regs68k));
            memset(regsppc, 0, sizeof(regsppc));

            regs68k[0] = src_data[src]; /* D0 */
            regs68k[1] = dst_data[dst]; /* D1 */
            regs68k[8] = dst_data[dst]; /* A0 */

            for (j=0; j<7; j++) { /* D0-D6, A0-A6 (D7 and A7 reserved) */
              regsppc[j+0] = regs68k[j+0];
              regsppc[j+8] = regs68k[j+8];
            }

            if (pre == 0) { /* All flags to zero */
              reg68k_sr.sr_int = 0x00;
            } else {        /* All flags to one */
              reg68k_sr.sr_int = 0x1f;
            }

            /* call */
            memset(&PP, 0, sizeof(struct PPCArgs));

            PP.PP_Code = (void *)buf68k;
            PP.PP_Offset = 0;
            PP.PP_Flags = 0;
            PP.PP_Stack = 0;
            PP.PP_StackSize = 0;

            for (j=0; j<7; j++) { /* D0-D6, A0-A6 (D7 and A7 reserved) */
              PP.PP_Regs[j+0] = regs68k[j+0];
              PP.PP_Regs[j+8] = regs68k[j+8];
            }

            PP.PP_Regs[PPREG_D7] = (ULONG)reg68k_sr.sr_int & 0x1f;

            Run68K(&PP);

            reg68k_sr.sr_int = PP.PP_Regs[PPREG_D7] & 0x1f;

            for (j=0; j<7; j++) { /* D0-D6, A0-A6 (D7 and A7 reserved) */
              regs68k[j+0] = PP.PP_Regs[j+0];
              regs68k[j+8] = PP.PP_Regs[j+8];
            }

            a_sr = reg68k_sr;

            if (pre == 0) {
              set_sr.sr_int |= a_sr.sr_int;
            } else {
              clr_sr.sr_int |= ~a_sr.sr_int;
            }

            if (pre == 0) { /* All flags to zero */
              reg68k_sr.sr_int = 0x00;
            } else {        /* All flags to one */
              reg68k_sr.sr_int = 0x1f;
            }

            /* call */
            op();

            b_sr = reg68k_sr;

#ifdef FLAGTEST
            if ((a_sr.sr_int & set) != (b_sr.sr_int & set)) {
              err = 1;
              printf("flag_error @ %2d,%2d,%2d,%2d,%2d (should be:  %s, is:  %s) (src %08x, dst %08x)\n", src, dst, iib->size, pre, set, showflags(a_sr), showflags(b_sr), src_data[src], dst_data[dst]);
            }
#endif
            for (j=0; j<16; j++) {
              if (regs68k[j] != regsppc[j]) {
                err = 1;
                printf("%s error  @ %2d,%2d,%2d,%2d,%2d (should be: %08x, is: %08x, flags:  %s)\n", regnames[j], src, dst, iib->size, pre, set, regs68k[j], regsppc[j], showflags(a_sr));
              }
            }
          }
        }
      }
    }

    set_sr.sr_int &= 0x1f;
    clr_sr.sr_int &= 0x1f;

    if (set_sr.sr_int != 0x1f) {
      set_sr.sr_int ^= 0x1f;
      printf("WARNING: flags %s were never set\n", showflags(set_sr));
    }

    if (clr_sr.sr_int != 0x1f) {
      clr_sr.sr_int ^= 0x1f;
      printf("WARNING: flags %s were never cleared\n", showflags(clr_sr));
    }

    if (err == 0) {
      printf("\tOK\n");
    } else {
      printf("\tFAILED\n");
    }

    i++;
  }
}

void optest(void)
{
  int i, src, dst, pre, set, s, size, err, a_vno, b_vno;
  uint32 a_srcdata, a_dstdata, b_srcdata, b_dstdata, *data, a_oldpc, b_oldpc;
  t_sr a_sr, b_sr, set_sr, clr_sr;
  t_iib iib;
  t_ipc ipc;
  op_t op;

  memset(&iib, 0, sizeof(t_iib));
  memset(&ipc, 0, sizeof(t_ipc));

  /* optests */
  i = 0;
  while (optests[i].name != NULL) {
    printf("testing %s...\n", optests[i].name);

    err = 0;
    set_sr.sr_int = 0;
    clr_sr.sr_int = 0;

    for (src=0; src<DATAS; src++) {
      for (dst=0; dst<DATAS; dst++) {
        for (s=0; s<3; s++) {
          size = optests[i].size[s];
          if (size == sz_none) {
            continue;
          }
          data = (size == sz_byte) ? byte_data : (size == sz_word) ? word_data : long_data;
          for (pre=0; pre<2; pre++) {
            for (set=0; set<2; set++) {
              if (pre == 0) { /* All flags to zero */
                TFLAG = SFLAG = XFLAG = NFLAG = ZFLAG = VFLAG = CFLAG = 0;
              } else {        /* All flags to one */
                TFLAG = SFLAG = XFLAG = NFLAG = ZFLAG = VFLAG = CFLAG = 1;
              }
              if (set == 0) { /* Set all flags */
                ipc.set = 0x1f;
              } else {        /* Set no flags */
                ipc.set = 0x00;
              }

              _vno = 0;
              _oldpc = 0;
              reg68k_pc = 0;
              srcdata = data[src];
              dstdata = data[dst];
              outdata = dstdata;
              _srcdata = srcdata;
              _dstdata = dstdata;
              iib.size = size;
              iib.mnemonic = optests[i].mnemonic;
              iib.wordlen = 2;

              /* call */
              optests[i].reference(&ipc, &iib);

              a_sr = reg68k_sr;
              a_srcdata = _srcdata;
              a_dstdata = _dstdata;
              a_vno = _vno;
              a_oldpc = _oldpc;

              if (pre == 0) { /* All flags to zero */
                set_sr.sr_int |= a_sr.sr_int;
              } else {        /* All flags to one */
                clr_sr.sr_int |= ~a_sr.sr_int;
              }

              if (pre == 0) { /* All flags to zero */
                TFLAG = SFLAG = XFLAG = NFLAG = ZFLAG = VFLAG = CFLAG = 0;
              } else {        /* All flags to one */
                TFLAG = SFLAG = XFLAG = NFLAG = ZFLAG = VFLAG = CFLAG = 1;
              }
              if (set == 0) { /* Set all flags */
                ipc.set = 0x1f;
              } else {        /* Set no flags */
                ipc.set = 0x00;
              }

              _vno = 0;
              _oldpc = 0;
              reg68k_pc = 0;
              srcdata = data[src];
              dstdata = data[dst];
              iib.size = size;
              iib.mnemonic = optests[i].mnemonic;
              iib.wordlen = 2;

              /* call */
              compile_op_init();
              compile_generic(&ipc, &iib);
              compile_op_exit();
              op = (op_t)buf;
              op();

              b_sr = reg68k_sr;
              b_srcdata = srcdata;
              b_dstdata = dstdata;
              b_vno = _vno;
              b_oldpc = _oldpc;

#ifdef FLAGTEST
              if (a_sr.sr_int != b_sr.sr_int) {
                err = 1;
                printf("flag_error @ %d,%d,%d,%d,%d (%s vs %s, %08x) (src %08x, dst %08x)\n", src, dst, size, pre, set, showflags(a_sr), showflags(b_sr), outdata, data[src], data[dst]);
              }
#endif
#ifdef SRCTEST
              if (a_srcdata != b_srcdata) {
                err = 1;
                printf("src_error @ %d,%d,%d,%d,%d (%08x vs %08x) (src %08x, dst %08x)\n", src, dst, size, pre, set, a_srcdata, b_srcdata, data[src], data[dst]);
              }
#endif
#ifdef DSTTEST
              if (a_dstdata != b_dstdata) {
                err = 1;
                printf("dst_error @ %d,%d,%d,%d,%d (%08x vs %08x) (src %08x, dst %08x)\n", src, dst, size, pre, set, a_dstdata, b_dstdata, data[src], data[dst]);
              }
#endif
#ifdef VECTORTEST
              if ((a_vno != b_vno) || (a_oldpc != b_oldpc)) {
                err = 1;
                printf("vector_error @ %d,%d,%d,%d,%d (%08x vs %08x, %08x vs %08x) (src %08x, dst %08x)\n", src, dst, size, pre, set, a_vno, b_vno, a_oldpc, b_oldpc, data[src], data[dst]);
              }
#endif
            }
          }
        }
      }
    }

    set_sr.sr_int &= 0x1f;
    clr_sr.sr_int &= 0x1f;

    if (set_sr.sr_int != 0x1f) {
      set_sr.sr_int ^= 0x1f;
      printf("WARNING: flags %s were never set\n", showflags(set_sr));
    }

    if (clr_sr.sr_int != 0x1f) {
      clr_sr.sr_int ^= 0x1f;
      printf("WARNING: flags %s were never cleared\n", showflags(clr_sr));
    }

    if (err == 0) {
      printf("\tOK\n");
    } else {
      printf("\tFAILED\n");
    }

    i++;
  }
}

int main(int argc, char **argv)
{
  compile_setup();

//  cctest();

  if (argc == 1) {
    printf("***** REAL68K *****\n");
    optest_real68k();
  } else {
    printf("***** INTERPRETER *****\n");
    optest();
  }

  compile_cleanup();

  return 0;
}