/* Generator is (c) James Ponder, 1997-2001 http://www.squish.net/generator/ */

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

#include "generator.h"
#include "gensound.h"
#include "gensoundp.h"
#include "vdp.h"
#include "ui.h"

#undef AUTOSKIP /* Define to get automatic frameskip */

#include <exec/types.h>
#include <exec/memory.h>
#include <devices/audio.h>

#include <powerup/ppcproto/intuition.h>
#include <powerup/ppclib/interface.h>
#include <powerup/ppclib/time.h>
#include <powerup/gcclib/powerup_protos.h>
#include <powerup/ppcproto/exec.h>
#include <powerup/ppcinline/alib.h>

#define AUDIO_CLOCK 3546895

char *final_wave = NULL, *final_wave_ptr = NULL;
struct IOAudio *AudioIO, *AudioIO1, *AudioIO2;
struct MsgPort *AudioMP, *AudioMP1, *AudioMP2;
int first_time = 2, audio_is_open = 0;
ULONG long device, wakebit;
BYTE *audioptr;
UBYTE whichannel[]={1, 2, 4, 8};
LONG samples;

/*** variables externed ***/

/*** forward references ***/

/*** file scoped variables ***/

static unsigned int soundp_block;
static int sample_buffer_size;

#ifdef AUTOSKIP
static struct timeval st = { 0, 0 };
#endif

/*** soundp_start - start sound hardware ***/

int soundp_start(void)
{
//  volatile unsigned char *filter = (unsigned char *)0xbfe001;
  audio_is_open = 0;

  /* Make sure that the variables are reset */
  device = 1;
  audioptr = NULL;
  AudioMP = NULL;
  AudioMP1 = NULL;
  AudioMP2 = NULL;
  AudioIO = NULL;
  AudioIO1 = NULL;
  AudioIO2 = NULL;

  soundp_block = 0;
  sample_buffer_size = sound_sampsperfield * sound_maxfields;

  final_wave = malloc(sample_buffer_size);
  final_wave_ptr = final_wave;

  /* Allocate resources */
  audioptr = (BYTE *)AllocMem(sample_buffer_size*2, MEMF_CHIP|MEMF_PUBLIC);
  if (audioptr != NULL)
  {
    AudioMP1 = CreateMsgPort();
    if (AudioMP1 != NULL)
    {
      AudioMP2 = CreateMsgPort();
      if (AudioMP2 != NULL)
      {
        AudioIO1 = CreateIORequest(AudioMP1, sizeof(struct IOAudio));
        if (AudioIO1 != NULL)
        {
          AudioIO2 = CreateIORequest(AudioMP2, sizeof(struct IOAudio));
          if (AudioIO2 != NULL)
          {
            AudioIO1->ioa_Request.io_Message.mn_ReplyPort   = AudioMP1;
            AudioIO1->ioa_Request.io_Message.mn_Node.ln_Pri = 127;
            AudioIO1->ioa_Request.io_Command                = ADCMD_ALLOCATE;
            AudioIO1->ioa_Request.io_Flags                  = ADIOF_NOWAIT;
            AudioIO1->ioa_AllocKey                          = 0;
            AudioIO1->ioa_Data                              = whichannel;
            AudioIO1->ioa_Length                            = sizeof(whichannel);

            device=OpenDevice(AUDIONAME, 0L, (struct IORequest *)AudioIO1, 0L);
            if (device == 0)
            {
              AudioIO1->ioa_Request.io_Command = CMD_WRITE;
              AudioIO1->ioa_Request.io_Flags   = ADIOF_PERVOL;
              AudioIO1->ioa_Volume             = 64;
              AudioIO1->ioa_Period             = AUDIO_CLOCK/sound_speed;
              AudioIO1->ioa_Cycles             = 1;
              AudioIO1->ioa_Length             = sample_buffer_size;

              /* Copy Settings */
              *AudioIO2 = *AudioIO1;

              /* Customize */
              AudioIO1->ioa_Data = (BYTE *)audioptr;
              AudioIO2->ioa_Data = (BYTE *)audioptr+sample_buffer_size;
              AudioIO1->ioa_Request.io_Message.mn_ReplyPort = AudioMP1;
              AudioIO2->ioa_Request.io_Message.mn_ReplyPort = AudioMP2;

//              printf("Audio is open\n");
              audio_is_open = 1;
              first_time = 2;
            }
          }
        }
      }
    }
  }

  if (audio_is_open != 1)
  {
    soundp_stop();
    return 1;
  }
  else
  {
    AudioIO = AudioIO1;
    AudioMP = AudioMP1;
  }

//  /* Turn off audio filter */
//  *filter |= 0x02;

//  /* Turn on audio filter */
//  *filter &= 0xfd;

#ifdef AUTOSKIP
  GetSysTimePPC(&st); /* Start */
#endif

  return 0;
}

/*** soundp_stop - stop sound hardware ***/

void soundp_stop(void)
{
  /* Abort any pending IO requests */
  if ((audio_is_open == 1) && (first_time == 0))
  {
    AbortIO((struct IORequest *)AudioIO1);
    WaitPort(AudioMP1);
    GetMsg(AudioMP1);
    AbortIO((struct IORequest *)AudioIO2);
    WaitPort(AudioMP2);
    GetMsg(AudioMP2);
  }

  /* Free resources */
  if (device == 0)
    CloseDevice((struct IORequest *)AudioIO1);
  if (AudioIO2 != NULL)
    DeleteIORequest(AudioIO2);
  if (AudioIO1 != NULL)
    DeleteIORequest(AudioIO1);
  if (AudioMP2 != NULL)
    DeleteMsgPort(AudioMP2);
  if (AudioMP1 != NULL)
    DeleteMsgPort(AudioMP1);
  if (audioptr != NULL)
    FreeMem(audioptr, sample_buffer_size*2);

//  if (audio_is_open == 1)
//    printf("Audio is closed\n");

  /* Make sure that the variables are reset */
  device = 1;
  audioptr = NULL;
  AudioMP = NULL;
  AudioMP1 = NULL;
  AudioMP2 = NULL;
  AudioIO = NULL;
  AudioIO1 = NULL;
  AudioIO2 = NULL;

  audio_is_open = 0;
  first_time = 2;
}

/*** soundp_samplesbuffered - how many samples are currently buffered? ***/

static int buffered = 0;

int soundp_samplesbuffered(void)
{
#ifdef AUTOSKIP
  double time;
  struct timeval dt = { 0, 0 };
  int played_samples, res;

  GetSysTimePPC(&dt);
  SubTimePPC(&dt, &st);

  time = (double)dt.tv_sec + ((double)dt.tv_usec / 1000000.0);
  played_samples = (int)((double)sound_speed * time);

  res = buffered - played_samples;
  if (res < 0) {
    res = 0;
  }

//
  printf("played_samples: %d, in buffer: %d (%d)\n", played_samples, res, buffered);
//

  return res;
#else
  return sound_threshold;
#endif
}

void soundp_output(uint16 *left, uint16 *right, unsigned int samples)
{
  int i;

  for (i=0; i<samples; i++) {
    *final_wave_ptr++ = left[i] >> 8;
  }

  buffered += samples;

  soundp_block++;
  if (soundp_block >= sound_maxfields) {
    soundp_block = 0;
    final_wave_ptr = final_wave;

    if (audio_is_open != 1)
      return;

    if (first_time == 0)
    {
      WaitPort(AudioMP);
      GetMsg(AudioMP);
    }
    else
    {
      first_time--;
#ifdef AUTOSKIP
      GetSysTimePPC(&st); /* Start */
      buffered = 0; /* Huh ? */
#endif
    }

    if (AudioIO == AudioIO1)
    {
      memcpy(audioptr, final_wave, sample_buffer_size);
      AudioIO->ioa_Request.io_Command = CMD_WRITE;
      AudioIO->ioa_Request.io_Flags   = ADIOF_PERVOL;
      AudioIO->ioa_Volume             = 64;
      AudioIO->ioa_Period             = AUDIO_CLOCK/sound_speed;
      AudioIO->ioa_Cycles             = 1;
      AudioIO->ioa_Length             = sample_buffer_size;
      AudioIO->ioa_Data               = (BYTE *)audioptr;
    }
    else
    {
      memcpy(audioptr+sample_buffer_size, final_wave, sample_buffer_size);
      AudioIO->ioa_Request.io_Command = CMD_WRITE;
      AudioIO->ioa_Request.io_Flags   = ADIOF_PERVOL;
      AudioIO->ioa_Volume             = 64;
      AudioIO->ioa_Period             = AUDIO_CLOCK/sound_speed;
      AudioIO->ioa_Cycles             = 1;
      AudioIO->ioa_Length             = sample_buffer_size;
      AudioIO->ioa_Data               = (BYTE *)audioptr+sample_buffer_size;
    }

    BeginIO((struct IORequest *)AudioIO);

    if (AudioIO == AudioIO1)
    {
      AudioIO = AudioIO2;
      AudioMP = AudioMP2;
    }
    else
    {
      AudioIO = AudioIO1;
      AudioMP = AudioMP1;
    }
  }
}
